@@ -15,7 +15,7 @@ module ParallelTests
       ENV['DISABLE_SPRING'] ||= '1'
 
       num_processes = ParallelTests.determine_number_of_processes(options[:count])
-      num_processes *= (options[:multiply] || 1)
+      num_processes = (num_processes * ParallelTests.determine_multiple(options[:multiply])).round
 
       options[:first_is_1] ||= first_is_1?
 
@@ -32,9 +32,23 @@ module ParallelTests
       @graceful_shutdown_attempted ||= false
       Kernel.exit if @graceful_shutdown_attempted
 
-      # The Pid class's synchronize method can't be called directly from a trap
-      # Using Thread workaround https://github.com/ddollar/foreman/issues/332
-      Thread.new { ParallelTests.stop_all_processes }
+      # In a shell, all sub-processes also get an interrupt, so they shut themselves down.
+      # In a background process this does not happen and we need to do it ourselves.
+      # We cannot always send the interrupt since then the sub-processes would get interrupted twice when in foreground
+      # and that messes with interrupt handling.
+      #
+      # (can simulate detached with `(bundle exec parallel_rspec test/a_spec.rb -n 2 &)`)
+      # also the integration test "passes on int signal to child processes" is detached.
+      #
+      # On windows getpgid does not work so we resort to always killing which is the smaller bug.
+      #
+      # The ParallelTests::Pids `synchronize` method can't be called directly from a trap,
+      # using Thread workaround https://github.com/ddollar/foreman/issues/332
+      Thread.new do
+        if Gem.win_platform? || ((child_pid = ParallelTests.pids.all.first) && Process.getpgid(child_pid) != Process.pid)
+          ParallelTests.stop_all_processes
+        end
+      end
 
       @graceful_shutdown_attempted = true
     end
@@ -43,8 +57,8 @@ module ParallelTests
       Tempfile.open 'parallel_tests-lock' do |lock|
         ParallelTests.with_pid_file do
           simulate_output_for_ci options[:serialize_stdout] do
-            Parallel.map(items, in_threads: num_processes) do |item|
-              result = yield(item)
+            Parallel.map_with_index(items, in_threads: num_processes) do |item, index|
+              result = yield(item, index)
               reprint_output(result, lock.path) if options[:serialize_stdout]
               ParallelTests.stop_all_processes if options[:fail_fast] && result[:exit_status] != 0
               result
@@ -67,8 +81,8 @@ module ParallelTests
         end
 
         report_number_of_tests(groups) unless options[:quiet]
-        test_results = execute_in_parallel(groups, groups.size, options) do |group|
-          run_tests(group, groups.index(group), num_processes, options)
+        test_results = execute_in_parallel(groups, groups.size, options) do |group, index|
+          run_tests(group, index, num_processes, options)
         end
         report_results(test_results, options) unless options[:quiet]
       end
@@ -82,8 +96,9 @@ module ParallelTests
       if any_test_failed?(test_results)
         warn final_fail_message
 
-        # return the highest exit status to allow sub-processes to send things other than 1
-        exit_status = if options[:highest_exit_status]
+        exit_status = if options[:failure_exit_code]
+          options[:failure_exit_code]
+        elsif options[:highest_exit_status]
           test_results.map { |data| data.fetch(:exit_status) }.max
         else
           1
@@ -94,8 +109,19 @@ module ParallelTests
     end
 
     def run_tests(group, process_number, num_processes, options)
-      if group.empty?
-        { stdout: '', exit_status: 0, command: nil, seed: nil }
+      if (limit = options[:test_file_limit])
+        # TODO: will have some bugs with summarizing results and last process
+        results = group.each_slice(limit).map do |slice|
+          @runner.run_tests(slice, process_number, num_processes, options)
+        end
+        result = results[0]
+        results[1..].each do |res|
+          result[:stdout] = result[:stdout].to_s + res[:stdout].to_s
+          result[:exit_status] = [res[:exit_status], result[:exit_status]].max
+          # adding all files back in, not using original cmd to show what was actually run
+          result[:command] |= res[:command]
+        end
+        result
       else
         @runner.run_tests(group, process_number, num_processes, options)
       end
@@ -136,7 +162,7 @@ module ParallelTests
         failing_sets.each do |failing_set|
           command = failing_set[:command]
           command = @runner.command_with_seed(command, failing_set[:seed]) if failing_set[:seed]
-          puts Shellwords.shelljoin(command)
+          @runner.print_command(command, failing_set[:env] || {})
         end
       end
     end
@@ -165,8 +191,9 @@ module ParallelTests
     end
 
     def parse_options!(argv)
-      newline_padding = " " * 37
+      newline_padding = 37 # poor man's way of getting a decent table like layout for -h output on 120 char width terminal
       options = {}
+
       OptionParser.new do |opts|
         opts.banner = <<~BANNER
           Run all tests in parallel, giving each process ENV['TEST_ENV_NUMBER'] ('', '2', '3', ...)
@@ -179,12 +206,14 @@ module ParallelTests
 
           Options are:
         BANNER
-        opts.on("-n [PROCESSES]", Integer, "How many processes to use, default: available CPUs") { |n| options[:count] = n }
-        opts.on("-p", "--pattern [PATTERN]", "run tests matching this regex pattern") { |pattern| options[:pattern] = /#{pattern}/ }
-        opts.on("--exclude-pattern", "--exclude-pattern [PATTERN]", "exclude tests matching this regex pattern") { |pattern| options[:exclude_pattern] = /#{pattern}/ }
+
+        opts.on("-n PROCESSES", Integer, "How many processes to use, default: available CPUs") { |n| options[:count] = n }
+        opts.on("-p", "--pattern PATTERN", "run tests matching this regex pattern") { |pattern| options[:pattern] = /#{pattern}/ }
+        opts.on("--exclude-pattern", "--exclude-pattern PATTERN", "exclude tests matching this regex pattern") { |pattern| options[:exclude_pattern] = /#{pattern}/ }
+
         opts.on(
-          "--group-by [TYPE]",
-          <<~TEXT.rstrip.split("\n").join("\n#{newline_padding}")
+          "--group-by TYPE",
+          heredoc(<<~TEXT, newline_padding)
             group tests by:
             found - order of finding files
             steps - number of cucumber/spinach steps
@@ -194,11 +223,12 @@ module ParallelTests
             default - runtime when runtime log is filled otherwise filesize
           TEXT
         ) { |type| options[:group_by] = type.to_sym }
-        opts.on("-m [FLOAT]", "--multiply-processes [FLOAT]", Float, "use given number as a multiplier of processes to run") do |multiply|
-          options[:multiply] = multiply
+
+        opts.on("-m COUNT", "--multiply-processes COUNT", Float, "use given number as a multiplier of processes to run") do |m|
+          options[:multiply_processes] = m
         end
 
-        opts.on("-s [PATTERN]", "--single [PATTERN]", "Run all matching files in the same process") do |pattern|
+        opts.on("-s PATTERN", "--single PATTERN", "Run all matching files in the same process") do |pattern|
           (options[:single_process] ||= []) << /#{pattern}/
         end
 
@@ -207,62 +237,102 @@ module ParallelTests
         end
 
         opts.on(
-          "--isolate-n [PROCESSES]",
+          "--isolate-n PROCESSES",
           Integer,
-          "Use 'isolate'  singles with number of processes, default: 1."
+          "Use 'isolate'  singles with number of processes, default: 1"
         ) { |n| options[:isolate_count] = n }
 
-        opts.on("--highest-exit-status", "Exit with the highest exit status provided by test run(s)") do
-          options[:highest_exit_status] = true
-        end
+        opts.on(
+          "--highest-exit-status",
+          "Exit with the highest exit status provided by test run(s)"
+        ) { options[:highest_exit_status] = true }
 
         opts.on(
-          "--specify-groups [SPECS]",
-          <<~TEXT.rstrip.split("\n").join("\n#{newline_padding}")
+          "--failure-exit-code INT",
+          Integer,
+          "Specify the exit code to use when tests fail"
+        ) { |code| options[:failure_exit_code] = code }
+
+        opts.on(
+          "--specify-groups SPECS",
+          heredoc(<<~TEXT, newline_padding)
             Use 'specify-groups' if you want to specify multiple specs running in multiple
             processes in a specific formation. Commas indicate specs in the same process,
-            pipes indicate specs in a new process. Cannot use with --single, --isolate, or
+            pipes indicate specs in a new process. If SPECS is a '-' the value for this
+            option is read from STDIN instead. Cannot use with --single, --isolate, or
             --isolate-n.  Ex.
-            $ parallel_test -n 3 . --specify-groups '1_spec.rb,2_spec.rb|3_spec.rb'
+            $ parallel_tests -n 3 . --specify-groups '1_spec.rb,2_spec.rb|3_spec.rb'
               Process 1 will contain 1_spec.rb and 2_spec.rb
               Process 2 will contain 3_spec.rb
               Process 3 will contain all other specs
           TEXT
         ) { |groups| options[:specify_groups] = groups }
 
-        opts.on("--only-group INT[,INT]", Array) { |groups| options[:only_group] = groups.map(&:to_i) }
+        opts.on(
+          "--only-group GROUP_INDEX[,GROUP_INDEX]",
+          Array,
+          heredoc(<<~TEXT, newline_padding)
+            Only run the given group numbers.
+            Changes `--group-by` default to 'filesize'.
+          TEXT
+        ) { |groups| options[:only_group] = groups.map(&:to_i) }
 
-        opts.on("-e", "--exec [COMMAND]", "execute this code parallel and with ENV['TEST_ENV_NUMBER']") { |arg| options[:execute] = Shellwords.shellsplit(arg) }
-        opts.on("-o", "--test-options '[OPTIONS]'", "execute test commands with those options") { |arg| options[:test_options] = Shellwords.shellsplit(arg) }
-        opts.on("-t", "--type [TYPE]", "test(default) / rspec / cucumber / spinach") do |type|
+        opts.on("-e", "--exec COMMAND", "execute COMMAND in parallel and with ENV['TEST_ENV_NUMBER']") { |arg| options[:execute] = Shellwords.shellsplit(arg) }
+        opts.on(
+          "--exec-args COMMAND",
+          heredoc(<<~TEXT, newline_padding)
+            execute COMMAND in parallel with test files as arguments, for example:
+            $ parallel_tests --exec-args echo
+            > echo spec/a_spec.rb spec/b_spec.rb
+          TEXT
+        ) { |arg| options[:execute_args] = Shellwords.shellsplit(arg) }
+
+        opts.on("-o", "--test-options 'OPTIONS'", "execute test commands with those options") { |arg| options[:test_options] = Shellwords.shellsplit(arg) }
+
+        opts.on("-t", "--type TYPE", "test(default) / rspec / cucumber / spinach") do |type|
           @runner = load_runner(type)
         rescue NameError, LoadError => e
           puts "Runner for `#{type}` type has not been found! (#{e})"
           abort
         end
+
         opts.on(
-          "--suffix [PATTERN]",
-          <<~TEXT.rstrip.split("\n").join("\n#{newline_padding}")
+          "--suffix PATTERN",
+          heredoc(<<~TEXT, newline_padding)
             override built in test file pattern (should match suffix):
-            '_spec\.rb$' - matches rspec files
+            '_spec.rb$' - matches rspec files
             '_(test|spec).rb$' - matches test or spec files
           TEXT
         ) { |pattern| options[:suffix] = /#{pattern}/ }
+
         opts.on("--serialize-stdout", "Serialize stdout output, nothing will be written until everything is done") { options[:serialize_stdout] = true }
         opts.on("--prefix-output-with-test-env-number", "Prefixes test env number to the output when not using --serialize-stdout") { options[:prefix_output_with_test_env_number] = true }
         opts.on("--combine-stderr", "Combine stderr into stdout, useful in conjunction with --serialize-stdout") { options[:combine_stderr] = true }
         opts.on("--non-parallel", "execute same commands but do not in parallel, needs --exec") { options[:non_parallel] = true }
         opts.on("--no-symlinks", "Do not traverse symbolic links to find test files") { options[:symlinks] = false }
-        opts.on('--ignore-tags [PATTERN]', 'When counting steps ignore scenarios with tags that match this pattern') { |arg| options[:ignore_tag_pattern] = arg }
+        opts.on('--ignore-tags PATTERN', 'When counting steps ignore scenarios with tags that match this pattern') { |arg| options[:ignore_tag_pattern] = arg }
         opts.on("--nice", "execute test commands with low priority.") { options[:nice] = true }
-        opts.on("--runtime-log [PATH]", "Location of previously recorded test runtimes") { |path| options[:runtime_log] = path }
-        opts.on("--allowed-missing [INT]", Integer, "Allowed percentage of missing runtimes (default = 50)") { |percent| options[:allowed_missing_percent] = percent }
-        opts.on("--unknown-runtime [FLOAT]", Float, "Use given number as unknown runtime (otherwise use average time)") { |time| options[:unknown_runtime] = time }
+        opts.on("--runtime-log PATH", "Location of previously recorded test runtimes") { |path| options[:runtime_log] = path }
+        opts.on("--allowed-missing COUNT", Integer, "Allowed percentage of missing runtimes (default = 50)") { |percent| options[:allowed_missing_percent] = percent }
+        opts.on('--allow-duplicates', 'When detecting files to run, allow duplicates') { options[:allow_duplicates] = true }
+        opts.on("--unknown-runtime SECONDS", Float, "Use given number as unknown runtime (otherwise use average time)") { |time| options[:unknown_runtime] = time }
         opts.on("--first-is-1", "Use \"1\" as TEST_ENV_NUMBER to not reuse the default test environment") { options[:first_is_1] = true }
         opts.on("--fail-fast", "Stop all groups when one group fails (best used with --test-options '--fail-fast' if supported") { options[:fail_fast] = true }
+
+        opts.on(
+          "--test-file-limit LIMIT",
+          Integer,
+          heredoc(<<~TEXT, newline_padding)
+            Limit to this number of files per test run by batching
+            (for windows set to ~100 to stay below 8192 max command limit, might have bugs from reusing test-env-number
+            and summarizing partial results)
+          TEXT
+        ) { |limit| options[:test_file_limit] = limit }
+
         opts.on("--verbose", "Print debug output") { options[:verbose] = true }
-        opts.on("--verbose-process-command", "Displays only the command that will be executed by each process") { options[:verbose_process_command] = true }
-        opts.on("--verbose-rerun-command", "When there are failures, displays the command executed by each process that failed") { options[:verbose_rerun_command] = true }
+        opts.on("--verbose-command", "Combines options --verbose-process-command and --verbose-rerun-command") { options.merge! verbose_process_command: true, verbose_rerun_command: true }
+        opts.on("--verbose-process-command", "Print the command that will be executed by each process before it begins") { options[:verbose_process_command] = true }
+        opts.on("--verbose-rerun-command", "After a process fails, print the command executed by that process") { options[:verbose_rerun_command] = true }
         opts.on("--quiet", "Print only tests output") { options[:quiet] = true }
         opts.on("-v", "--version", "Show Version") do
           puts ParallelTests::VERSION
@@ -306,22 +376,26 @@ module ParallelTests
         raise "--group-by #{allowed.join(" or ")} is required for --only-group"
       end
 
-      if options[:specify_groups] && (options.keys & [:single_process, :isolate, :isolate_count]).any?
+      if options[:specify_groups] && options.keys.intersect?([:single_process, :isolate, :isolate_count])
         raise "Can't pass --specify-groups with any of these keys: --single, --isolate, or --isolate-n"
       end
 
+      if options[:failure_exit_code] && options[:highest_exit_status]
+        raise "Can't pass --failure-exit-code and --highest-exit-status"
+      end
+
       options
     end
 
     def extract_file_paths(argv)
       dash_index = argv.rindex("--")
       file_args_at = (dash_index || -1) + 1
-      [argv[file_args_at..-1], argv[0...(dash_index || 0)]]
+      [argv[file_args_at..], argv[0...(dash_index || 0)]]
     end
 
     def extract_test_options(argv)
       dash_index = argv.index("--") || -1
-      argv[dash_index + 1..-1]
+      argv[dash_index + 1..]
     end
 
     def append_test_options(options, argv)
@@ -404,5 +478,9 @@ module ParallelTests
         yield
       end
     end
+
+    def heredoc(text, newline_padding)
+      text.rstrip.gsub("\n", "\n#{' ' * newline_padding}")
+    end
   end
 end
